#include <stdio.h>
#include <fcntl.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>

#include "M6502.h"

byte VIA_read(word addr);
void VIA_write(word addr,byte data);
int VIA_poll_int();
void VIA_reset();

byte LCD_read(word addr);
void LCD_write(word addr,byte data);

M6502 regs;

byte RAM[4096];
byte RAM9[4096]=
#include "a1asm.h"
byte RAME[4096]=
#include "a1basic.h"
byte ROM[256]=
#include "wozmon.h"
byte ROMCIA[256];

byte keyboard=0;
byte video;

byte cpu_readmem(word addr)
{
     unsigned int a=addr&0xfff;
     byte t;
     
     switch(addr>>12){
     case 0:	return RAM[a];
     case 9:	return RAM9[a];
     case 0xe:	return RAME[a];
     case 0xd:  a=addr&3;
     		switch(a) {
		case 0: t=keyboard;
			keyboard&=0x7f;
			return t;
		case 1: return keyboard&0x80;
		case 2: return video;
		case 3: return 0;
		}
     case 0xf:	return ROM[addr&0xff];
     default:	return 0xff;
     }
}

void cpu_writemem(word addr,byte data)
{
     unsigned int a=addr&0xfff;
     byte t;
     
     switch(addr>>12){
     case 0:	RAM[a]=data; break;
     case 9:	RAM9[a]=data; break;
     case 0xe:	RAME[a]=data; break;
     case 0xd:  a=addr&3;
     		switch(a) {
		case 2:	video=data|0x80; break;
		}
     }
}

unsigned long maxclocks=0;

// this is not a real interrupt, but an emulation interrupt:
//   the emulation stops and these tasks are carried out periodically
// a better name would be something like "periodic_update" ...
//
// Called every regs.IPeriod clocks to poll interrupt status and 
// to emulate I/O from the PC side

int cpu_interrupt()
{
    int i,j,irq=INT_NONE;
    static int cntv=0;
    unsigned char a;
    
    // Read stdin (keyboard emulation)
    if (read(0,&a,1)==1) {
    	if (a=='\n') a='\r';	   // LF translated as CR
	if (a=='X'-64) a='C'-64;   // ctrl-X tranaslated as ctrl-C
	if (a==0x7f) a='_';	   // backspace translation
	keyboard=toupper(a)|0x80;
    }
    // Video ACK emulation (worst case: 16.7 ms write delay)
    if (++cntv==100) {
        cntv=0;
	if (video&0x80) {
           video&=0x7f;	// Clear flag
	   a=video;
	   if (a=='\r') a='\n';	// CR translated as LF
	   if (a=='_') a='\b';	// backspace translation
	   write(1,&a,1);
	}
    }
    // Pause emulation after 40 ms of emulated time
    if((maxclocks-regs.Clocks)&0x80000000) {
        pause();
    }
    
    return irq;
}

void trace_rti()
{
        //fprintf(stderr,"%d\n",regs.Clocks-irqstamp);
}


void alarm_40m() // Called every 40 ms
{
    maxclocks+=(CLKFREQ/25);
}

void ctrlc()	// CTRL - C is pressed => enter trace mode
{
    regs.Trace=1;
}

/******************************************************************************

		M	A	I	N

******************************************************************************/
int stdinflags;
extern FILE *dskp;

main(int argc, char **argv)
{
        
	int fd;
	unsigned int i;
        struct itimerval t0,t1;
        struct sigaction act0,act1;


	
        for (i=0;i<0x1000;i++) cpu_writemem(i,i>>8);


        t0.it_interval.tv_sec=0;
        t0.it_interval.tv_usec=40000;
        t0.it_value=t0.it_interval;

        act0.sa_handler=alarm_40m;
        act0.sa_flags=SA_RESTART;

        sigaction(SIGALRM,&act0,&act1);
        setitimer(ITIMER_REAL,&t0,&t1);
	
	signal(SIGINT,ctrlc);

        // Start the emulation
        Reset6502(&regs);
	
	stdinflags=fcntl(0, F_GETFL, 0);
	
	regs.IPeriod=167; // video write every 100 cpu_interrupts
        regs.Trace=1;	  // enter debugger at startup
        Run6502(&regs);
}

